/**
 * -- Digitalis Utilities Project --
 *  - Inspection Utilities -
 *
 * (C) 2007, Digitalis Informatica.
 *
 * Distribuicao e Gestao de Informatica, Lda.
 * Estrada de Paco de Arcos num.9 - Piso -1
 * 2780-666 Paco de Arcos
 * Telefone: (351) 21 4408990
 * Fax: (351) 21 4408999
 * http://www.digitalis.pt
 */

package pt.digitalis.utils.inspection;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.StringTokenizer;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;

import pt.digitalis.utils.inspection.exception.ResourceNotFoundException;

/**
 * This class provides classpath searching tools for classes and files. It provides an abstraction of the Java classpath
 * making easier the access to class path resources.
 *
 * @author Pedro Viegas <a href="mailto:pviegas@digitalis.pt">pviegas@digitalis.pt</a>
 * @author Rodrigo Gonalves <a href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a>
 * @created Oct 9, 2007
 */
final public class ClasspathUtils {

    /** The system property class path key. */
    final static String JAVA_CLASS_PATH_SYSTEM_PROPERTY = "java.class.path";

    /**
     * Defines the behavior of a filter whose filtering criteria is file-name based.
     */
    public interface FilenameFilter {
        /**
         * Defines the filtering criteria.
         *
         * @param filename
         *            the file name
         * @return T if filename matches the criteria, F otherwise
         */
        public boolean accept(String filename);
    }

    /** Enforce noninstantiability. */
    private ClasspathUtils() {
    }

    /**
     * Returns an array of the classpath classes names. The method's return consists the class's FQN without the .class
     * file extension, thus making it useful for object instantiation with the java.lang.Class.forName(String) method.
     *
     * @return a String array with the classpath classes names
     * @throws ResourceNotFoundException
     *             if a Jar file can't be found or properly opened
     */
    public static List<String> getClasspathClassNames() throws ResourceNotFoundException {

        List <String> classes = getClasspathFileNamesWithExtension(".class");

        for (String clazz : classes)
            clazz.substring(0, clazz.length() - 6).replace("/", ".");

        return classes;
    }

    /**
     * Returns an array of the classpath files ending with a given extension/filename suffix.
     *
     * @param extension
     *            the file extension/suffix
     * @return an array of the classpath files that match the criteria mentioned above
     * @throws ResourceNotFoundException
     *             if a Jar file can't be found or properly opened
     */
    public static List<String> getClasspathFileNamesWithExtension(final String extension)
            throws ResourceNotFoundException {
        return getClasspathFileNames(new FilenameFilter() {
            public boolean accept(String filename) {
                return filename.endsWith(extension);
            }
        });
    }

    /**
     * Returns an array of the classpath file names that match the given file name filter.
     *
     * @param filter
     *            the filter object
     * @return an array of the classpath file names matching the applied criteria
     * @throws ResourceNotFoundException
     *             if a Jar file can't be found or properly opened
     */
    public static List<String> getClasspathFileNames(FilenameFilter filter) throws ResourceNotFoundException {

        final List<String> filenames = new ArrayList<String>();

        for (String filename : getClasspathFileNames()) {
            if (filter.accept(filename)) {
                filenames.add(filename);
            }
        }
        return filenames;
    }

    /**
     * Returns the FQNs of all the classes on the classpath. Checks directories and Jar files.
     *
     * @return a list of the classpath classes names
     * @throws ResourceNotFoundException
     *             if a Jar file can't be found or properly opened
     */
    public static List<String> getClasspathFileNames() throws ResourceNotFoundException {

        final StringTokenizer tokenizer = new StringTokenizer(System.getProperty(JAVA_CLASS_PATH_SYSTEM_PROPERTY),
                File.pathSeparator, false);
        final Set<String> filenames = new LinkedHashSet<String>();

        while (tokenizer.hasMoreTokens()) {

            final String classpathElement = tokenizer.nextToken();
            final File classpathFile = new File(classpathElement);

            if (classpathFile.exists() && classpathFile.canRead()) {
                if (classpathElement.toLowerCase(Locale.US).endsWith(".jar")) {

                    ZipFile zip = null;
                    try {
                        zip = new ZipFile(classpathFile);
                    } catch (ZipException zipException) {
                        throw new ResourceNotFoundException("Jar file couldn't be opened!", zipException);
                    } catch (IOException ioException) {
                        throw new ResourceNotFoundException("Jar file couldn't be found!", ioException);
                    }

                    Enumeration<? extends ZipEntry> entries = zip.entries();

                    while (entries.hasMoreElements()) {
                        ZipEntry entry = entries.nextElement();
                        if (!entry.isDirectory()) {
                            filenames.add(entry.getName());
                        }
                    }

                } else if (classpathFile.isDirectory()) {
                    // Recursively process subfolders
                    final Set<File> directoriesToSearch = new HashSet<File>();
                    final Set<File> newDirectories = new HashSet<File>();
                    directoriesToSearch.add(classpathFile);
                    final String basePath = classpathFile.getAbsolutePath();

                    while (directoriesToSearch.size() > 0) {
                        for (File searchDirectory : directoriesToSearch) {
                            File[] directoryFiles = searchDirectory.listFiles();
                            for (File directoryFile : directoryFiles) {
                                if (directoryFile.isDirectory()) {
                                    newDirectories.add(directoryFile);
                                } else {
                                    filenames.add(directoryFile.getAbsolutePath().substring(basePath.length() + 1));
                                }
                            }
                        }
                        directoriesToSearch.clear();
                        directoriesToSearch.addAll(newDirectories);
                        newDirectories.clear();
                    }
                }
            }
        }

        final String[] uniqueNames = new String[filenames.size()];
        int index = 0;

        for (String name : filenames) {
            uniqueNames[index++] = name.replace("\\", "/");
        }

        return Arrays.asList(uniqueNames);
    }
}
