/**
 * -- Digitalis Utilities Project -- - Inspection Utilities - (C) 2007, Digitalis Informatica. Distribuicao e Gestao de
 * Informatica, Lda. Estrada de Paco de Arcos num.9 - Piso -1 2780-666 Paco de Arcos Telefone: (351) 21 4408990 Fax:
 * (351) 21 4408999 http://www.digitalis.pt
 */
package pt.digitalis.utils.inspection;

import java.lang.annotation.Annotation;
import java.lang.annotation.ElementType;
import java.lang.annotation.Target;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import pt.digitalis.utils.inspection.exception.AuxiliaryOperationException;
import pt.digitalis.utils.inspection.exception.ResourceNotFoundException;

/**
 * This class presents a set of methods that provide common Java reflection needs, such as reading annotations, methods
 * and fields for a given class. There also methods to instantiate objects and invoke methods. This wrapper was designed
 * to cope with the Java's reflection method and API dispersion. This API is still being developed and the future might
 * show the need to refactor it to interfaces and implementations. This option may prove itself useful if, for example,
 * a need to use different annotation processing mechanism arises.
 * 
 * @author Rodrigo Gonalves <a href="mailto:rgoncalves@digitalis.pt">rgoncalves@digitalis.pt</a>
 * @created 2007/05/15
 */
final public class ReflectionUtils {

    /**
     * Returns a list with all annotations for a given class. All annotation scopes are considered (Type, Method and
     * Attribute).
     * 
     * @param clazz
     *            the FQN of the class to look for attribute annotations
     * @return a list with all the class's annotation names
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getAllAnnotationsForClass(Class<?> clazz) throws ResourceNotFoundException
    {

        // Init result
        List<Annotation> allAnnotations = new ArrayList<Annotation>();

        // Init annotation buffer...and get type annotations for the class
        List<Annotation> annotations = getTypeAnnotationsForClass(clazz);

        // Process fetched annotations (check if annotation already exists)
        for (Annotation annotation: annotations)
        {
            if (!allAnnotations.contains(annotation))
                allAnnotations.add(annotation);
        }

        // Get method annotations for class...
        annotations = getMethodAnnotationsForClass(clazz);

        // Process fetched annotations (check if annotation already exists)
        for (Annotation annotation: annotations)
        {
            if (!allAnnotations.contains(annotation))
                allAnnotations.add(annotation);
        }
        // Get attribute annotations for class...
        annotations = getAttributeAnnotationsForClass(clazz);

        // Process fetched annotations (check if annotation already exists)
        for (Annotation annotation: annotations)
        {
            if (!allAnnotations.contains(annotation))
                allAnnotations.add(annotation);
        }

        return allAnnotations;
    }

    /**
     * Returns a list with all annotations for a given class. All annotation scopes are considered (Type, Method and
     * Attribute).
     * 
     * @param className
     *            the FQN of the class to look for attribute annotations
     * @return a list with all the class's annotation names
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getAllAnnotationsForClass(String className) throws ResourceNotFoundException
    {

        Class<?> clazz = null;

        try
        {
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getAllAnnotationsForClass(clazz);
    }

    /**
     * Returns the defined target scopes for a given annotation.
     * 
     * @param annotation
     *            the annotation to be read.
     * @return a list of allowed target scopes for the annotation
     */
    static public List<ElementType> getAnnotationTarget(Annotation annotation)
    {

        Annotation[] annotationAnnotations = annotation.annotationType().getDeclaredAnnotations();

        List<ElementType> targetScopes = null;

        for (Annotation annotationAnnotation: annotationAnnotations)
        {
            if (annotationAnnotation instanceof Target)
            {
                Target targetMetaAnnotation = (Target) annotationAnnotation;

                targetScopes = Arrays.asList(targetMetaAnnotation.value());
            }
        }

        return targetScopes;
    }

    /**
     * Returns a given attribute on a given class.
     * 
     * @param clazz
     *            the class object
     * @param attributeName
     *            the name of the attribute to fetch
     * @return the attribute with the given name if it's found, null otherwise
     */
    static public Field getAttribute(Class<?> clazz, String attributeName)
    {

        for (Field attribute: getAttributes(clazz))
        {
            if (attributeName.equals(attribute.getName()))
                return attribute;
        }

        return null;
    }

    /**
     * Searches a class with a given name and returns the attribute with the given name, if it exists.
     * 
     * @param className
     *            the class's name
     * @param attributeName
     *            the attribute's name
     * @return the attribute with the given name if it's found, null otherwise
     * @throws ResourceNotFoundException
     *             if the class with the given name can't be found
     */
    static public Field getAttribute(String className, String attributeName) throws ResourceNotFoundException
    {
        Class<?> clazz = null;

        try
        {
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getAttribute(clazz, attributeName);
    }

    /**
     * Returns a list with a given class attribute annotations.
     * 
     * @param clazz
     *            the class to look for attribute annotations
     * @return a list with the class's attribute annotations
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getAttributeAnnotationsForClass(Class<?> clazz) throws ResourceNotFoundException
    {

        // Fetch methods from class
        List<Field> fields = Arrays.asList(clazz.getDeclaredFields());

        // Init return
        List<Annotation> attributesAnnotations = new ArrayList<Annotation>();

        // For each attribute...
        for (Field field: fields)
        {

            // ...fetch it's annotations...
            List<Annotation> attributeAnnotations = Arrays.asList(field.getDeclaredAnnotations());

            // ...for each annotation...
            for (Annotation annotation: attributeAnnotations)
            {
                // ...add it to result if it's not there yet...
                if (!attributesAnnotations.contains(annotation))
                    attributesAnnotations.add(annotation);
            }
        }

        return attributesAnnotations;
    }

    /**
     * Returns a list with a given class attribute annotations.
     * 
     * @param className
     *            the FQN of the class to look for attribute annotations
     * @return a list with the class's attribute annotations
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getAttributeAnnotationsForClass(String className) throws ResourceNotFoundException
    {

        // Init class holder
        Class<?> clazz = null;

        try
        {
            // Get class object
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getAttributeAnnotationsForClass(clazz);
    }

    /**
     * Return a list with a given class attributes.
     * 
     * @param clazz
     *            the class to get the attributes
     * @return a list with the class fields
     */
    static public List<Field> getAttributes(Class<?> clazz)
    {
        return Arrays.asList(clazz.getDeclaredFields());
    }

    /**
     * Return a list with a given class attributes.
     * 
     * @param className
     *            the name of the class to get the attributes
     * @return a list with the class fields
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Field> getAttributes(String className) throws ResourceNotFoundException
    {
        Class<?> clazz = null;
        try
        {
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getAttributes(clazz);
    }

    /**
     * Searches a given class and returns the method with the given name, if it exists.
     * 
     * @param clazz
     *            the class object
     * @param methodName
     *            the method's name
     * @return the method with the given name if it's found, null otherwise
     */
    static public Method getMethod(Class<?> clazz, String methodName)
    {

        for (Method method: getMethods(clazz))
            if (methodName.equals(method.getName()))
                return method;

        return null;
    }

    /**
     * Searches a class with a given name and returns the method with the given name, if it exists.
     * 
     * @param className
     *            the class's name
     * @param methodName
     *            the method's name
     * @return the method with the given name if it's found, null otherwise
     * @throws ResourceNotFoundException
     *             if the class with the given name can't be found
     */
    static public Method getMethod(String className, String methodName) throws ResourceNotFoundException
    {
        Class<?> clazz = null;

        try
        {
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getMethod(clazz, methodName);
    }

    /**
     * Returns a list with the annotations from all methods on a given class.
     * 
     * @param clazz
     *            the class to process
     * @return a list with all methods annotations
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getMethodAnnotationsForClass(Class<?> clazz) throws ResourceNotFoundException
    {

        // Fetch methods from class
        Method[] methods = clazz.getDeclaredMethods();

        // Init return
        List<Annotation> methodsAnnotations = new ArrayList<Annotation>();

        // For each method...
        for (Method method: methods)
        {

            // Fetch it's annotations
            List<Annotation> methodAnnotations = Arrays.asList(method.getDeclaredAnnotations());

            // ...and for each annotation...
            for (Annotation methodAnnotation: methodAnnotations)
            {
                // ...add it to result if it's not already there...
                if (!methodsAnnotations.contains(methodAnnotation))
                    methodsAnnotations.add(methodAnnotation);
            }
        }

        return methodsAnnotations;
    }

    /**
     * Returns a list with the annotations from all methods on a given class.
     * 
     * @param className
     *            the name of the class to process
     * @return a list with all methods annotations
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getMethodAnnotationsForClass(String className) throws ResourceNotFoundException
    {

        // Init class holder
        Class<?> clazz = null;

        try
        {
            // Get class object
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getMethodAnnotationsForClass(clazz);
    }

    /**
     * Searches a given class method and returns a list of the parameter names. Requires the use of {@link Named}
     * annotation in each parameter
     * 
     * @param clazz
     *            the class object
     * @param methodName
     *            the method's name
     * @return the list of parameter names
     * @throws NoSuchMethodException
     */
    static public List<String> getMethodParameterNames(Class<?> clazz, String methodName) throws NoSuchMethodException
    {
        Method[] methods = clazz.getMethods();

        for (Method method: methods)
            if (method.getName().equals(methodName))
                return getMethodParameterNames(method);

        throw new NoSuchMethodException(methodName);
    }

    /**
     * Searches a given class method and returns a list of the parameter names. Requires the use of {@link Named}
     * annotation in each parameter
     * 
     * @param method
     *            the method to inspect
     * @return the list of parameter names
     */
    static public List<String> getMethodParameterNames(Method method)
    {
        List<String> parameterNames = new ArrayList<String>();

        for (Annotation[] parameterAnnotations: method.getParameterAnnotations())
        {
            String name = null;

            for (Annotation annotation: parameterAnnotations)
                if (annotation instanceof Named)
                {
                    name = ((Named) annotation).value();
                    break;
                }

            parameterNames.add(name);
        }

        return parameterNames;
    }

    /**
     * Return a list with a given class methods.
     * 
     * @param clazz
     *            the class to get the methods
     * @return a list with the class methods
     */
    static public List<Method> getMethods(Class<?> clazz)
    {
        return Arrays.asList(clazz.getMethods());
    }

    /**
     * Return a list with a given class methods.
     * 
     * @param className
     *            the name of the class to get the attributes
     * @return a list with the class methods
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Method> getMethods(String className) throws ResourceNotFoundException
    {
        Class<?> clazz = null;
        try
        {
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException notFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    notFoundException);
        }

        return getMethods(clazz);
    }

    /**
     * Returns a list with a given class annotations.
     * 
     * @param clazz
     *            the class to look for annotations
     * @return a list with the class annotations
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getTypeAnnotationsForClass(Class<?> clazz) throws ResourceNotFoundException
    {

        Annotation[] fetchedAnnotations = clazz.getDeclaredAnnotations();

        return Arrays.asList(fetchedAnnotations);
    }

    /**
     * Returns a list with a given class annotations.
     * 
     * @param className
     *            the FQN of the class to look for annotations
     * @return a list with the class annotations
     * @throws ResourceNotFoundException
     *             if class cannot be found.
     */
    static public List<Annotation> getTypeAnnotationsForClass(String className) throws ResourceNotFoundException
    {

        // Init class holder
        Class<?> clazz = null;

        try
        {
            // Get class object
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException classNotFoundException)
        {
            throw new ResourceNotFoundException("Class " + className + " was not found and thus could not be loaded!",
                    classNotFoundException);
        }

        return getTypeAnnotationsForClass(clazz);
    }

    /**
     * Instantiates an object given it's class.
     * 
     * @param <T>
     *            the class type
     * @param clazz
     *            the class to instantiate
     * @return an instance of the given class
     * @throws AuxiliaryOperationException
     *             if it's not possible to instantiate an object from the given class
     */
    static public <T> T instantiateObjectFromClass(Class<T> clazz) throws AuxiliaryOperationException
    {

        try
        {
            return clazz.newInstance();
        }
        catch (InstantiationException instantiationException)
        {
            throw new AuxiliaryOperationException("Could not instantiate object from class " + clazz.getName() + "(Is "
                    + clazz.getName() + " an interface or an abstract class?)", instantiationException);
        }
        catch (IllegalAccessException illegalAccessException)
        {
            throw new AuxiliaryOperationException("Could not instantiate object from class " + clazz.getName() + "(Is "
                    + clazz.getName() + " constructor private?)", illegalAccessException);
        }
    }

    /**
     * Instantiates an object given it's constructor.
     * 
     * @param <T>
     *            the class type
     * @param constructor
     *            the constructor for the class to instantiate
     * @param arguments
     *            the constructor's argument
     * @return an instance of the given class
     * @throws AuxiliaryOperationException
     *             if it's not possible to instantiate an object from the given class
     */
    static public <T> T instantiateObjectFromClass(Constructor<T> constructor, Object... arguments)
            throws AuxiliaryOperationException
    {

        try
        {
            return constructor.newInstance(arguments);

        }
        catch (InstantiationException e)
        {
            throw new AuxiliaryOperationException("Could not instantiate object from constructor "
                    + constructor.getName(), e);
        }
        catch (IllegalAccessException e)
        {
            throw new AuxiliaryOperationException("Could not instantiate object from class " + constructor.getName(), e);
        }
        catch (IllegalArgumentException e)
        {
            throw new AuxiliaryOperationException("Could not instantiate object from constructor "
                    + constructor.getName() + ": Illegal argument for no arguments passed", e);
        }
        catch (InvocationTargetException e)
        {
            throw new AuxiliaryOperationException("Could not instantiate object from constructor "
                    + constructor.getName(), e);
        }
    }

    /**
     * Invokes a method on a given object.
     * 
     * @param method
     *            the method to invoke
     * @param object
     *            the invocation target object
     * @param arguments
     *            the called method's arguments
     * @return the executed method's return
     * @throws AuxiliaryOperationException
     *             if either an illegal argument is passed to the method, the method hasn't public access or if the
     *             target object throws an exception
     */
    static public Object invokeMethod(Method method, Object object, Object... arguments)
            throws AuxiliaryOperationException
    {

        try
        {
            return method.invoke(object, arguments);
        }
        catch (IllegalArgumentException illegalArgumentException)
        {
            throw new AuxiliaryOperationException(illegalArgumentException);
        }
        catch (IllegalAccessException illegalAccessException)
        {
            throw new AuxiliaryOperationException(illegalAccessException);
        }
        catch (InvocationTargetException invocationTargetException)
        {
            throw new AuxiliaryOperationException(invocationTargetException.getCause().getMessage(),
                    invocationTargetException);
        }
    }

    /**
     * Overloaded version of <code>invokeMethod(..)</code> with String-type parameters.
     * 
     * @param methodName
     *            the name of the method to invoke
     * @param className
     *            the name of the invocation target object
     * @param arguments
     *            the called method's arguments
     * @return the executed method's return
     * @throws AuxiliaryOperationException
     *             if the class can't be loaded or any of the reasons in
     *             pt.digitalis.utils.inspection#invokeMethod(Method, Object, Object...)
     * @see pt.digitalis.utils.inspection#invokeMethod(Method, Object, Object...)
     */
    static public Object invokeMethod(String methodName, String className, Object... arguments)
            throws AuxiliaryOperationException
    {

        Object result = null;

        try
        {
            Method method = getMethod(className, methodName);

            if (method == null)
                throw new AuxiliaryOperationException("Could not found method " + methodName + " on class " + className);

            result = invokeMethod(method, instantiateObjectFromClass(Class.forName(className)), arguments);

        }
        catch (AuxiliaryOperationException auxiliaryOperationException)
        {
            throw auxiliaryOperationException;
        }
        catch (ResourceNotFoundException resourceNotFoundException)
        {
            throw new AuxiliaryOperationException("Could not load class " + className
                    + " and thus could not found method " + methodName, resourceNotFoundException);
        }
        catch (ClassNotFoundException classNotFoundException)
        {
            throw new AuxiliaryOperationException("Could not load class " + className
                    + " and thus could not found method " + methodName, classNotFoundException);
        }

        return result;
    }

    /**
     * Returns T if a given class has a given annotation. This method captures type, method and attribute annotations.
     * 
     * @param className
     *            the class to search for the annotation
     * @param annotationName
     *            the name of the annotation to search
     * @return T if the class is annotated with the annotation, F otherwise
     * @throws ResourceNotFoundException
     *             if the class can't be found
     */
    static public boolean isAnnotatedWith(String className, String annotationName) throws ResourceNotFoundException
    {
        Class<?> clazz;

        try
        {
            // Get class
            clazz = Class.forName(className);
        }
        catch (ClassNotFoundException classNotFoundException)
        {
            throw new ResourceNotFoundException("Could not found class " + className + "!", classNotFoundException);
        }

        // Get class annotations
        List<Annotation> annotations = getAllAnnotationsForClass(clazz);

        // Search for the given annotation
        for (Annotation annotation: annotations)
        {
            if (annotation.annotationType().getCanonicalName().equals(annotationName))
                return true;
        }

        return false;
    }

    /**
     * Checks if a given method is an inspector. Inspector are nullary functions with names started with "get" and
     * returning anything but void.
     * 
     * @param method
     *            the method to test
     * @return T if its a simple getter method, F otherwise
     */
    static public boolean isInspectorMethod(Method method)
    {
        return (method.getName().startsWith("get") && (method.getGenericParameterTypes().length == 0) && !method
                .getReturnType().equals("void"));
    }

    /** Enforce noninstantiability. */
    private ReflectionUtils()
    {}
}
